#include <iostream>
#include <Windows.h>
#include "igstkArucoTracker.h"
#include "igstkTransformObserver.h"


/**
 * This program uses a monocular (read webcam) tracker to send "Page Down" key
 * and "Page Up" key to either a user specified program that will have focus or
 * the current program that has focus. In practice this is a free remote control 
 * for a power point presentation. Allows the user to move forward and backward 
 * in the presentation by showing tracked markers (see the markers.pdf file).
 *
 * Page Down:          marker 202
 * Page Up:            marker 502
 * Exit this program:  marker 201.
 * 
 * To build this program from source you will need to:
 * a. build the Insight Segmentation and Registration Toolkit (ITK) [www.itk.org]
 * b. build the Visualization Toolkit (VTK) [www.vtk.org]
 * c. build the Open Source Computer Vision library (OpenCV) [www.opencv.org]
 * d. build the Image-Guided Surgery Toolkit (IGSTK) [www.igstk.org]
 *
 * Copyright Ziv Yaniv (ZYaniv@childrensnational.org)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define igstkEventOccurredMacro( name, EventType ) \
class name##EventOccurredObserver : public ::itk::Command \
{ \
public: \
  typedef  name##EventOccurredObserver  Self; \
  typedef  ::itk::Command               Superclass;\
  typedef  ::itk::SmartPointer<Self>    Pointer;\
  itkNewMacro( Self );\
public:\
  void Execute(itk::Object *caller, const itk::EventObject & event)\
    {\
    const itk::Object * constCaller = caller;\
    this->Execute( constCaller, event );\
    }\
  void Execute(const itk::Object *caller, const itk::EventObject & event)\
    {\
    if( EventType().CheckEvent( &event ) )\
      {\
        m_EventOccurred = true;\
      }\
    }\
  bool EventOccured() const\
    {\
    return m_EventOccurred;\
    }\
  void Reset() \
    {\
    m_EventOccurred = false; \
    }\
protected:\
  name##EventOccurredObserver() \
    {\
    m_EventOccurred = false;\
    }\
  ~name##EventOccurredObserver() {}\
private:\
  bool m_EventOccurred;\
};


igstkEventOccurredMacro( TrackerOpen,  igstk::TrackerOpenEvent )
igstkEventOccurredMacro( StartTracking, igstk::TrackerStartTrackingEvent )


/*
 * The GenerateKey function was obtained here:
 * http://stackoverflow.com/questions/5202118/c-troubleshooting-sending-keystrokes
 *
 * Apparently this is not the original, and it is not clear who is the original 
 * author, so kudos go to the anonymous coder.
 */
void GenerateKey( int vk, BOOL bExtended );


/**
 * This is where we setup the tracking and associate between tools and actions.
 * THIS IS THE HEART OF THE PROGRAM.
 */
igstk::Tracker::Pointer Setup();


/**
 * Observer class for the ToolUpdated event which terminates the program, stops
 * the tracking, closes the connection to the tracker and exits.
 */
class ToolUpdatedTerminateProgramObserver  : public ::itk::Command
{
public:
  typedef  ToolUpdatedTerminateProgramObserver    Self;
  typedef  ::itk::Command                         Superclass;
  typedef  ::itk::SmartPointer<Self>              Pointer;
  itkNewMacro( Self );

protected:

  ToolUpdatedTerminateProgramObserver() 
  {
    this->m_TransformObserver = igstk::TransformObserver::New();    
  }

  ~ToolUpdatedTerminateProgramObserver() {}

public:
  void Initialize( igstk::TrackerTool::Pointer trackerTool, 
		           igstk::Tracker::Pointer tracker ) 
   {
     this->m_Tool = trackerTool;
     this->m_TransformObserver->ObserveTransformEventsFrom( this->m_Tool );
     this->m_LastValid = 0;
	 this->m_Tracker = tracker;
   }

  void Execute( itk::Object *caller, const itk::EventObject & event )
  {
    const itk::Object * constCaller = caller;
    this->Execute( constCaller, event );
  }
  
  void Execute( const itk::Object *caller, const itk::EventObject & event )
  {
    const double DURATION_VISIBLE = 2000.0; //time in milliseconds

                 //do something only for the correct tool
    if( this->m_Tool.GetPointer() == caller )
    {               //the tool transform has been updated, get it
      if( dynamic_cast<const
            igstk::TrackerToolTransformUpdateEvent  *>( &event) )
      {                 //request to get the transform
        this->m_Tool->RequestGetTransformToParent();
        if ( this->m_TransformObserver->GotTransform() )
        {
          igstk::Transform transform =
            this->m_TransformObserver->GetTransform();
          transform.GetStartTime();
          igstk::TransformBase::TimePeriodType curStart = 
            transform.GetStartTime();
          if( curStart - this->m_LastValid > DURATION_VISIBLE ) {
			this->m_Tracker->RequestStopTracking();
			this->m_Tracker->RequestClose();
			exit( EXIT_SUCCESS );
          }          
        }
      }
    }
  }

private:
  igstk::TransformObserver::Pointer m_TransformObserver; 
  igstk::TrackerTool::Pointer  m_Tool;
  igstk::TransformBase::TimePeriodType m_LastValid;
  igstk::Tracker::Pointer m_Tracker;
};


/**
 * Observer class for the ToolUpdated event which sends a specific key event
 * to the window that currently has focus.
 */
class ToolUpdatedGenerateKeyEventObserver : public ::itk::Command
{
public:
  typedef  ToolUpdatedGenerateKeyEventObserver    Self;
  typedef  ::itk::Command                         Superclass;
  typedef  ::itk::SmartPointer<Self>              Pointer;
  itkNewMacro( Self );

protected:

  ToolUpdatedGenerateKeyEventObserver() 
  {
    this->m_TransformObserver = igstk::TransformObserver::New();    
  }

  ~ToolUpdatedGenerateKeyEventObserver() {}

 public:
   void Initialize( igstk::TrackerTool::Pointer trackerTool, int vk ) 
   {
     this->m_Tool = trackerTool;
     this->m_TransformObserver->ObserveTransformEventsFrom( this->m_Tool );
     this->m_LastValid = 0;
	 this->m_vk = vk;
   }

  void Execute( itk::Object *caller, const itk::EventObject & event )
  {
    const itk::Object * constCaller = caller;
    this->Execute( constCaller, event );
  }
  
  void Execute( const itk::Object *caller, const itk::EventObject & event )
  {
    const double DURATION_VISIBLE = 1000.0; //time in milliseconds

                 //do something only for the correct tool
    if( this->m_Tool.GetPointer() == caller )
    {               //the tool transform has been updated, get it
      if( dynamic_cast<const
            igstk::TrackerToolTransformUpdateEvent  *>( &event) )
      {                 //request to get the transform
        this->m_Tool->RequestGetTransformToParent();
        if ( this->m_TransformObserver->GotTransform() )
        {
          igstk::Transform transform =
            this->m_TransformObserver->GetTransform();
          transform.GetStartTime();
          igstk::TransformBase::TimePeriodType curStart = 
            transform.GetStartTime();
          if( curStart - this->m_LastValid > DURATION_VISIBLE ) {
			GenerateKey( this->m_vk, false );
            this->m_LastValid = curStart;
          }          
        }
      }
    }
  }

private:
  igstk::TransformObserver::Pointer m_TransformObserver; 
  igstk::TrackerTool::Pointer  m_Tool;
  igstk::TransformBase::TimePeriodType m_LastValid;
  int m_vk;
};



/**
 * Generate a key event which is sent to the window that currently has focus.
 * @param vk Virtual key code. These codes are defined in WinUser.h .
 */
void GenerateKey( int vk, BOOL bExtended )
{
  KEYBDINPUT  kb = {0};
  INPUT input = {0};

    //generate a "key down"
  if ( bExtended ) 
    kb.dwFlags  = KEYEVENTF_EXTENDEDKEY;
  kb.wVk  = vk;
  input.type  = INPUT_KEYBOARD;
  input.ki  = kb;
  SendInput( 1, &input, sizeof( INPUT ) );

    //generate a "key up"
  ZeroMemory( &kb, sizeof( KEYBDINPUT ) );
  ZeroMemory( &input, sizeof( INPUT ) );
  kb.dwFlags  =  KEYEVENTF_KEYUP;
  if ( bExtended ) 
    kb.dwFlags |= KEYEVENTF_EXTENDEDKEY;
  kb.wVk = vk;
  input.type = INPUT_KEYBOARD;
  input.ki = kb;
  SendInput( 1, &input, sizeof( INPUT ) );

  return;
}

/**
 * Configure the ArucoTracker. Attach three tools corresponding to the forward
 * keypress, backward keypress, and exit actions. The former key events are sent to the 
 * window that has focus. The exit action will terminate this program after
 * shutting the tracker down.
 */
igstk::Tracker::Pointer Setup()
{                   //camera calibration file should be in the same directory
	                //in which the program is run
  const std::string CAMERA_CALIBRATION_FILENAME = "cameraCalibration.yml";
                  //marker size in mm
  const unsigned int MARKER_SIZE = 50;
                 //requested tracking refresh rate in Hz
  const double REFRESH_RATE = 5;

  const std::string FORWARD_MARKER_NAME = "202";
  const std::string BACKWARD_MARKER_NAME = "502";
  const std::string EXIT_MARKER_NAME = "201";

  igstk::Tracker::Pointer tracker;
  igstk::ArucoTrackerTool::Pointer forwardTool, backwardTool, terminateTool;

  igstk::ArucoTracker::Pointer  arucoTracker = igstk::ArucoTracker::New();
  arucoTracker->SetCameraParametersFromYAMLFile( CAMERA_CALIBRATION_FILENAME );
  arucoTracker->SetMarkerSize( MARKER_SIZE );
  arucoTracker->RequestSetFrequency( REFRESH_RATE );
             //need to observe that request open worked
  TrackerOpenEventOccurredObserver::Pointer too = TrackerOpenEventOccurredObserver::New();
  arucoTracker->AddObserver( igstk::TrackerOpenEvent(), too );
  arucoTracker->RequestOpen();

                    //if the tracker open succeeded we add the tools and attempt to
                    //start tracking
  if( too->EventOccured() ) {
	           //setup tool for sending the page-down/forward event
    forwardTool = igstk::ArucoTrackerTool::New();
    forwardTool->RequestSetMarkerName( FORWARD_MARKER_NAME );
    forwardTool->RequestConfigure();
    forwardTool->RequestAttachToTracker( arucoTracker );
  
    ToolUpdatedGenerateKeyEventObserver::Pointer forwardObserver = 
		ToolUpdatedGenerateKeyEventObserver::New();
    forwardObserver->Initialize( forwardTool.GetPointer(), VK_NEXT );
    forwardTool->AddObserver( igstk::TrackerToolTransformUpdateEvent(), forwardObserver );

	          //setup tool for sending the page-up/backward event
    backwardTool = igstk::ArucoTrackerTool::New();
    backwardTool->RequestSetMarkerName( BACKWARD_MARKER_NAME );
    backwardTool->RequestConfigure();
    backwardTool->RequestAttachToTracker( arucoTracker );
  
    ToolUpdatedGenerateKeyEventObserver::Pointer backwardObserver = 
		ToolUpdatedGenerateKeyEventObserver::New();
    backwardObserver->Initialize( backwardTool.GetPointer(), VK_PRIOR );
    backwardTool->AddObserver( igstk::TrackerToolTransformUpdateEvent(), backwardObserver );
	
	          //setup tool for terminating the program
    terminateTool = igstk::ArucoTrackerTool::New();
    terminateTool->RequestSetMarkerName( EXIT_MARKER_NAME );
    terminateTool->RequestConfigure();
    terminateTool->RequestAttachToTracker( arucoTracker );
  
    ToolUpdatedTerminateProgramObserver::Pointer terminateObserver = 
		ToolUpdatedTerminateProgramObserver::New();
	terminateObserver->Initialize( terminateTool.GetPointer(), arucoTracker.GetPointer() );
    terminateTool->AddObserver( igstk::TrackerToolTransformUpdateEvent(), terminateObserver );

            //attempt to start tracking
    StartTrackingEventOccurredObserver::Pointer sto = StartTrackingEventOccurredObserver::New();
    arucoTracker->AddObserver( igstk::TrackerStartTrackingEvent(), sto );
    arucoTracker->RequestStartTracking();
    if( sto->EventOccured() )
      tracker = arucoTracker;
    else 
      std::cerr<<"Failed to start tracker.\n";
  }
  else {
   std::cerr<<"Failed to open tracker [camera not connected, or calibration file not found].\n";
  } 
  return tracker;
}


/*
 * Example of running the program: 
 * pptcontrol "tracking.pptx - Microsoft PowerPoint"
 */
int main( int argc, char *argv[] )
{
       //send keystrokes to the target window specified by the user, otherwise
	   //they are sent to the program that has focus
  if( argc == 2 ) {
    HWND targetWindow;
          //find the user specified window and set the focus to this window
    targetWindow = FindWindow( NULL, argv[1] );
    if( targetWindow == NULL ) {
      std::cerr<<"Could not find specified target window.\n";
      return EXIT_FAILURE;
    }
        //set the focus to this top level window, all key events will be sent
        //here
    SetForegroundWindow( targetWindow );
  }

  igstk::RealTimeClock::Initialize();

  try 
  {
         //get the tracker so that it is in the program scope
    igstk::Tracker::Pointer tracker = Setup();
    if( tracker.IsNull() )
      return EXIT_FAILURE;
    
    while( true ) {
      igstk::PulseGenerator::Sleep( 10 );
      igstk::PulseGenerator::CheckTimeouts();
    }
    return EXIT_SUCCESS;
  }
  catch( std::exception &e )
  {
    std::cerr<<e.what()<<"\n";
    return EXIT_FAILURE;
  }
}



